from __future__ import annotations

from functools import wraps
from typing import Any, Callable, Generic, Optional, TypeVar, Union, cast

import numpy as np
from numpy.typing import NDArray
from scipy.sparse import csc_array  # type: ignore
from typing_extensions import Concatenate, ParamSpec

from ._typing import DecayingLearner

P = ParamSpec("P")
R = TypeVar("R", covariant=True)
RewardFunction = Union[
    Callable[..., np.float_],
    Callable[..., NDArray[np.float_]],
    Callable[..., Union[np.float_, NDArray[np.float_]]],
]
LT = TypeVar("LT", bound=DecayingLearner)
A = TypeVar("A", bound="Arm[Any, Any]")
T = TypeVar("T")


def requires_learner(
    func: Callable[Concatenate[A, P], R],
) -> Callable[Concatenate[A, P], R]:
    """Decorator to check if the arm has a learner set."""

    @wraps(func)
    def wrapper(self: A, *args: P.args, **kwargs: P.kwargs) -> R:
        if self.learner is None:  # type: ignore
            raise ValueError("Learner is not set.")
        return func(self, *args, **kwargs)

    return wrapper


def identity(
    x: Union[np.float_, NDArray[np.float_]],
) -> Union[np.float_, NDArray[np.float_]]:
    return x


class Arm(Generic[LT, T]):
    """Arm of a bandit.

    Parameters
    ----------
    action_token : Any
        Token to return when the arm is pulled. This should be something processed
        by the user's code to execute the action associated with the arm.
    reward_function : Optional[RewardFunction], default=None
        Function to call to compute the reward. Takes the output of the learner's
        `sample` function as input and should return a scalar reward. Should take
        a single scalar or array-like argument and return a scalar or
        1D array. If None, the identity function is used.
    learner : Optional[Learner], default=None
        Learner to use for the arm. If None, the arm cannot be used.

    Examples
    --------
    >>> from bayesianbandits import Arm
    >>> import numpy as np
    >>> def reward_function(sample):
    ...     return sample
    >>> class MyLearner:
    ...     def sample(self, X, size=1):
    ...         np.random.seed(0)
    ...         return np.random.normal(size=size)
    ...     def partial_fit(self, X, y):
    ...         pass
    >>> learner = MyLearner()
    >>> arm = Arm("Action taken.", reward_function, learner)
    >>> arm.pull()
    'Action taken.'
    >>> arm.update(np.array([[1]]), np.array([1]))

    """

    def __init__(
        self,
        action_token: T,
        reward_function: Optional[RewardFunction] = None,
        learner: Optional[LT] = None,
    ) -> None:
        self.action_token: T = action_token
        if reward_function is None:
            reward_function = identity
        self.reward_function = reward_function
        self.learner: LT = learner  # type: ignore

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    @requires_learner
    def pull(self) -> T:
        """Pull the arm."""
        return self.action_token

    @requires_learner
    def sample(
        self,
        X: Union[csc_array, NDArray[np.float_]],
        size: int = 1,
    ) -> NDArray[np.float_]:
        """Sample from learner and compute the reward."""
        return self.reward_function(self.learner.sample(X, size))  # type: ignore

    @requires_learner
    def update(
        self, X: Union[csc_array, NDArray[np.float_]], y: NDArray[np.float_]
    ) -> None:
        """Update the learner.

        If y is None, the data in X is used as the target and X is set to
        a `len(X)` rows of ones.
        """
        assert self.learner is not None  # for type checker
        # sparse learners are supported by some, but not all learners
        self.learner.partial_fit(X, y)  # type: ignore

    @requires_learner
    def decay(
        self,
        X: Union[csc_array, NDArray[np.float_]],
        *,
        decay_rate: Optional[float] = None,
    ) -> None:
        """Decay the learner.

        Takes a `y` argument for consistency with `update` but does not use it."""
        if not hasattr(self.learner, "decay"):
            raise ValueError("Learner does not have a decay method.")
        # sparse learners are supported by some, but not all learners
        cast(DecayingLearner, self.learner).decay(X, decay_rate=decay_rate)  # type: ignore

    def __repr__(self) -> str:
        return (
            f"Arm(action_token={self.action_token},"
            f" reward_function={self.reward_function}"
        )



from __future__ import annotations

from typing import (
    Any,
    NewType,
    Optional,
    Protocol,
    Union,
)

import numpy as np
from numpy.typing import NDArray


ActionToken = NewType("TokenType", Any)


class Learner(Protocol):
    """Learner protocol for the model underlying each arm.

    Each Learner must implement the following methods:
    - `sample`
    - `partial_fit`

    """

    random_state: Union[np.random.Generator, int, None]

    def sample(
        self,
        X: NDArray[Any],
        size: int = 1,
    ) -> NDArray[np.float_]:
        ...

    def partial_fit(self, X: NDArray[Any], y: NDArray[Any]) -> "Learner":
        ...

    def set_params(self, **params: Any) -> "Learner":
        ...


class DecayingLearner(Learner, Protocol):
    learning_rate: float

    def decay(self, X: NDArray[Any], *, decay_rate: Optional[float] = None) -> None:
        ...



"""
=======================================================
Agent API (:mod:`bayesianbandits.api`)
=======================================================

.. currentmodule:: bayesianbandits.api

Fully typed API for Bayesian bandits.

This module contains the API for Bayesian bandits. It passes strict type
checking with `pyright` and is recommended for use in production code. On top
of the type checking, this API makes it much easier to add or remove arms, change
the policy function, and serialize/deserialize bandits in live services.

It uses the same estimators and `Arm` class as the original API, but
defines its own `ContextualAgent` and `Agent` classes, as well as a 
`Policy` type alias for the policy functions.

This API splits the `Bandit` class into two classes, `ContextualAgent`
and `Agent`. This split enables safer typing, as the contextual bandit
always takes a context matrix as input, while the non-contextual bandit does not.

Additionally, this API deprecates the `delayed_reward` decorator, as it modifies
the function signatures of the `pull` and `update` methods. Instead, this API
enables batch pulls and updates, but leaves it up to the user to keep track of
matching updates with the correct pulls. Library users reported that this was
what they were doing anyway, so this API change should encourage better practices.

.. note::

    Migrating from the original API to this API should be straightforward. Just
    instantiate the `ContextualAgent` or `MultiArmedBandit` class
    with `list(arms.values())` and the policy function of the original `Bandit`
    subclass.

Bandit Classes
==============

.. autosummary::
    :toctree: _autosummary

    ContextualAgent
    Agent

Policy Functions
================

.. autosummary::
    :toctree: _autosummary

    EpsilonGreedy
    ThompsonSampling
    UpperConfidenceBound

"""


from typing import (
    Any,
    Callable,
    Generic,
    List,
    Optional,
    Set,
    TypeVar,
    Union,
    cast,
)

import numpy as np
from numpy.typing import NDArray
from scipy.sparse import csc_array  # type: ignore
from typing_extensions import Self

from ._arm import Arm
from ._basebandit import _validate_arrays  # type: ignore
from ._policy_decorators import (
    _compute_arm_mean,  # type: ignore
    _compute_arm_upper_bound,  # type: ignore
    _draw_one_sample,  # type: ignore
)
from ._typing import DecayingLearner

AT = TypeVar("AT", bound=Arm[Any, Any])
T = TypeVar("T")
L = TypeVar("L", bound=DecayingLearner)

Policy = Callable[
    [List[Arm[Any, Any]], Union[NDArray[np.float_], csc_array], np.random.Generator],
    List[Arm[Any, Any]],
]


class ContextualAgent(Generic[L, T]):
    """Agent for a contextual multi-armed bandit problem.

    Parameters
    ----------
    arms : List[Arm]
        List of arms to choose from. All arms must have a learner and a unique
        action token.
    policy : Callable[[List[Arm], NDArray[np.float_], Generator], Arm]
        Function to choose an arm from the list of arms. Takes the list of arms
        and the context as input and returns the chosen arm.
    random_seed : int, default=None
        Seed for the random number generator. If None, a random seed is used.

    Attributes
    ----------
    arms : List[Arm]
        List of arms to choose from. All arms must have a learner and a unique
        action token.
    policy : Callable[[List[Arm], NDArray[np.float_], Generator], Arm]
        Function to choose an arm from the list of arms. Takes the list of arms
        and the context as input and returns the chosen arm.
    arm_to_update : Arm
        Arm to update with the next reward.
    rng : Generator
        Random number generator used for choosing arms and decaying.


    Examples
    --------

    Minimally, an agent can be instantiated with a list of arms and a policy
    function. The arms should have a learner and a unique action token.

    >>> from bayesianbandits import Arm, NormalInverseGammaRegressor
    >>> from bayesianbandits.api import ContextualAgent, ThompsonSampling
    >>> arms = [
    ...     Arm(0, learner=NormalInverseGammaRegressor()),
    ...     Arm(1, learner=NormalInverseGammaRegressor()),
    ... ]
    >>> agent = ContextualAgent(arms, ThompsonSampling(), random_seed=0)

    The agent can then be used to choose an arm and pull it. The `pull` method
    takes a context matrix as input and returns the action token of the chosen
    arm.

    >>> import numpy as np
    >>> X = np.array([[1.0, 15.0]])
    >>> agent.pull(X)
    [0]

    The agent can then be updated with the observed reward. The `update` method
    takes a context matrix and a reward vector as input. By default, the last
    pulled arm is updated.

    >>> y = np.array([100.0])
    >>> agent.update(X, y)
    >>> agent.arm_to_update.learner.predict(X)
    array([99.55947137])

    The agent can also be 'fluently' updated by chaining the `arm` method with
    the `update` method. This is useful when the reward is not immediately
    available, and a batch of updates needs to be made later. The user is
    trusted with calling `arm` with the right action token.

    >>> agent.select_for_update(1).update(X, y)
    >>> agent.arm_to_update is arms[1]
    True
    >>> agent.arm_to_update.learner.predict(X)
    array([99.55947137])
    """

    def __repr__(self) -> str:
        learners = set(type(arm.learner) for arm in self.arms)
        action_space = set(arm.action_token for arm in self.arms)
        reward_function = set(arm.reward_function for arm in self.arms)

        return (
            f"ContextualAgent(policy={self.policy}, random_seed={self.rng},\n"
            f"arms={action_space}, reward_function={reward_function},\n"
            f"learners={learners})"
        )

    def __init__(
        self,
        arms: List[Arm[L, T]],
        policy: Policy,
        random_seed: Union[int, None, np.random.Generator] = None,
    ):
        self._arms = arms
        if not len(arms) > 0:
            raise ValueError("At least one arm is required.")
        unique_tokens: Set[T] = set(arm.action_token for arm in arms)
        if not len(unique_tokens) == len(arms):
            raise ValueError("All arms must have unique action tokens.")
        if not all(arm.learner is not None for arm in arms):  # type: ignore
            raise ValueError("All arms must have a learner.")

        self.policy: Callable[
            [
                List[Arm[L, T]],
                Union[NDArray[np.float_], csc_array],
                np.random.Generator,
            ],
            List[Arm[L, T]],
        ] = policy

        self.arm_to_update = arms[0]

        self.rng: np.random.Generator = np.random.default_rng(random_seed)
        for arm in self.arms:
            arm.learner.random_state = random_seed

    @property
    def arms(self) -> List[Arm[L, T]]:
        return self._arms

    def add_arm(self, arm: Arm[L, T]) -> None:
        """Add an arm to the bandit.

        Parameters
        ----------
        arm : Arm
            Arm to add to the bandit.

        Raises
        ------
        ValueError
            If the arm's action token is already in the bandit.
        """
        current_tokens = set(arm.action_token for arm in self.arms)
        if arm.action_token in current_tokens:
            raise ValueError("All arms must have unique action tokens.")
        self.arms.append(arm)

    def remove_arm(self, token: T) -> None:
        """Remove an arm from the bandit.

        Parameters
        ----------
        token : Any
            Action token of the arm to remove.

        Raises
        ------
        KeyError
            If the arm's action token is not in the bandit.
        """
        for i, arm in enumerate(self._arms):
            if arm.action_token == token:
                self._arms.pop(i)
                break
        else:
            raise KeyError(f"Arm with token {token} not found.")

    def arm(self, token: T) -> Arm[L, T]:
        """Get an arm by its action token.

        Parameters
        ----------
        token : Any
            Action token of the arm to get.

        Returns
        -------
        Arm
            Arm with the action token.

        Raises
        ------
        KeyError
            If the arm's action token is not in the bandit.
        """
        for arm in self.arms:
            if arm.action_token == token:
                return arm
        raise KeyError(f"Arm with token {token} not found.")

    def select_for_update(self, token: T) -> Self:
        """Set the `arm_to_update` and return self for chaining.

        Parameters
        ----------
        token : Any
            Action token of the arm to update.

        Returns
        -------
        Self
            Self for chaining.

        Raises
        ------
        KeyError
            If the arm's action token is not in the bandit.
        """

        for arm in self.arms:
            if arm.action_token == token:
                self.arm_to_update = arm
                break
        else:
            raise KeyError(f"Arm with token {token} not found.")
        return self

    def pull(self, X: Union[NDArray[np.float_], csc_array]) -> List[T]:
        """Choose an arm and pull it based on the context(s).

        Parameters
        ----------
        X : Union[NDArray[np.float_], csc_array]
            Context matrix to use for choosing an arm.

        Returns
        -------
        List[ActionToken]
            List of action tokens for the pulled arms.
        """
        X_pull, _ = _validate_arrays(X, None, contextual=True, check_y=False)
        arms = self.policy(self.arms, X_pull, self.rng)
        self.arm_to_update = arms[-1]
        return [arm.pull() for arm in arms]

    def update(
        self, X: Union[NDArray[np.float_], csc_array], y: NDArray[np.float_]
    ) -> None:
        """Update the `arm_to_update` with the context(s) and the reward(s).

        Parameters
        ----------
        X : Union[NDArray[np.float_], csc_array]
            Context matrix to use for updating the arm.
        y : NDArray[np.float_]
            Reward(s) to use for updating the arm.
        """
        X_updated, y_update = _validate_arrays(X, y, contextual=True, check_y=True)
        self.arm_to_update.update(X_updated, y_update)

    def decay(
        self,
        X: Union[NDArray[np.float_], csc_array],
        decay_rate: Optional[float] = None,
    ) -> None:
        """Decay all arms of the bandit len(X) times.

        Parameters
        ----------
        X : Union[NDArray[np.float_], csc_array]
            Context matrix to use for decaying the arm.
        decay_rate : Optional[float], default=None
            Decay rate to use for decaying the arm. If None, the decay rate
            of the arm's learner is used.
        """
        X_decay, _ = _validate_arrays(X, None, contextual=True, check_y=False)
        for arm in self.arms:
            arm.decay(X_decay, decay_rate=decay_rate)


class Agent(Generic[L, T]):
    """
    Agent for a non-contextual multi-armed bandit problem.

    The non-contextual bandit is a special case of the contextual bandit where
    the context matrix is a single column of ones. This class is a wrapper
    around the `ContextualAgent` class that automatically synthesizes the
    context matrix for the `pull`, `update`, and `decay` methods.

    Parameters
    ----------
    arms : List[Arm]
        List of arms to choose from. All arms must have a learner and a unique
        action token.
    policy : Callable[[List[Arm], NDArray[np.float_], Generator], Arm]
        Function to choose an arm from the list of arms. Takes the list of arms
        and the context as input and returns the chosen arm.
    random_seed : int, default=None
        Seed for the random number generator. If None, a random seed is used.

    Attributes
    ----------
    arms : List[Arm]
        List of arms to choose from. All arms must have a learner and a unique
        action token.
    policy : Callable[[List[Arm], NDArray[np.float_], Generator], Arm]
        Function to choose an arm from the list of arms. Takes the list of arms
        and the context as input and returns the chosen arm.
    arm_to_update : Arm
        Arm to update with the next reward.

    Examples
    --------
    The key difference between the `Agent` and the `ContextualAgent` is that
    the `Agent` does not take a context matrix as input. Instead, the context
    matrix is synthesized automatically.

    >>> from bayesianbandits import Arm, NormalInverseGammaRegressor
    >>> from bayesianbandits.api import Agent, ThompsonSampling
    >>> arms = [
    ...     Arm(0, learner=NormalInverseGammaRegressor()),
    ...     Arm(1, learner=NormalInverseGammaRegressor()),
    ... ]
    >>> agent = Agent(arms, ThompsonSampling(), random_seed=0)
    >>> agent.pull()
    [0]

    This is equivalent to calling the `pull` method with a context matrix containing
    only a global intercept. The `update` and `decay` methods work the same way.

    >>> import numpy as np
    >>> y = np.array([100.0])
    >>> agent.update(y)
    >>> agent.select_for_update(0).update(y)
    """

    def __repr__(self) -> str:
        learners = set(type(arm.learner) for arm in self.arms)
        action_space = set(arm.action_token for arm in self.arms)
        reward_function = set(arm.reward_function for arm in self.arms)

        return (
            f"Agent(policy={self.policy}, random_seed={self.rng},\n"
            f"arms={action_space}, reward_function={reward_function},\n"
            f"learners={learners})"
        )

    def __init__(
        self,
        arms: List[Arm[L, T]],
        policy: Policy,
        random_seed: Union[int, None, np.random.Generator] = None,
    ):
        self._inner: ContextualAgent[L, T] = ContextualAgent(
            arms, policy, random_seed=random_seed
        )

    @property
    def policy(self):
        return self._inner.policy

    @policy.setter
    def policy(
        self,
        policy: Callable[
            [
                List[Arm[L, T]],
                Union[NDArray[np.float_], csc_array],
                np.random.Generator,
            ],
            List[Arm[L, T]],
        ],
    ) -> None:
        self._inner.policy = policy

    @property
    def rng(self) -> np.random.Generator:
        return self._inner.rng

    @property
    def arm_to_update(self) -> Arm[L, T]:
        return self._inner.arm_to_update

    @property
    def arms(self) -> List[Arm[L, T]]:
        return self._inner.arms

    def add_arm(self, arm: Arm[L, T]) -> None:
        """Add an arm to the bandit.

        Parameters
        ----------
        arm : Arm
            Arm to add to the bandit.

        Raises
        ------
        ValueError
            If the arm's action token is already in the bandit.
        """
        self._inner.add_arm(arm)

    def remove_arm(self, token: Any) -> None:
        """Remove an arm from the bandit.

        Parameters
        ----------
        token : Any
            Action token of the arm to remove.

        Raises
        ------
        KeyError
            If the arm's action token is not in the bandit.
        """
        self._inner.remove_arm(token)

    def select_for_update(self, token: T) -> Self:
        """Set the `arm_to_update` and return self for chaining.

        Parameters
        ----------
        token : Any
            Action token of the arm to update.

        Returns
        -------
        Self
            Self for chaining.

        Raises
        ------
        KeyError
            If the arm's action token is not in the bandit.
        """
        self._inner.select_for_update(token)
        return self

    def arm(self, token: T) -> Arm[L, T]:
        """Get an arm by its action token.

        Parameters
        ----------
        token : Any
            Action token of the arm to get.

        Returns
        -------
        Self
            Self for chaining.

        Raises
        ------
        KeyError
            If the arm's action token is not in the bandit.
        """
        return self._inner.arm(token)

    def pull(self):
        """Choose an arm and pull it.

        Returns
        -------
        List[ActionToken]
            List containing the action token for the pulled arm.
        """
        X_pull, _ = _validate_arrays(None, None, contextual=False, check_y=False)
        return self._inner.pull(X_pull)

    def update(self, y: NDArray[np.float_]) -> None:
        """Update the `arm_to_update` with an observed reward.

        Parameters
        ----------
        y : NDArray[np.float_]
            Reward(s) to use for updating the arm.
        """
        X_update, y_update = _validate_arrays(y, None, contextual=False, check_y=True)
        self._inner.update(X_update, y_update)

    def decay(self, decay_rate: Optional[float] = None) -> None:
        """Decay all arms of the bandit.

        Parameters
        ----------
        decay_rate : Optional[float], default=None
            Decay rate to use for decaying the arm. If None, the decay rate
            of the arm's learner is used.
        """
        X_decay, _ = _validate_arrays(None, None, contextual=False, check_y=False)
        self._inner.decay(X_decay, decay_rate=decay_rate)


class EpsilonGreedy:
    """
    Policy object for epsilon-greedy.

    Parameters
    ----------
    epsilon : float, default=0.1
        Probability of exploration.
    samples : int, default=1000
        Number of samples to use for computing the arm means.

    """

    def __repr__(self) -> str:
        return f"EpsilonGreedy(epsilon={self.epsilon}, samples={self.samples})"

    def __init__(self, epsilon: float = 0.1, samples: int = 1000):
        self.epsilon = epsilon
        self.samples = samples

    def __call__(
        self,
        arms: List[Arm[L, T]],
        X: Union[NDArray[np.float_], csc_array],
        rng: np.random.Generator,
    ) -> List[Arm[L, T]]:
        """Choose an arm using epsilon-greedy."""
        means = np.stack(
            tuple(_compute_arm_mean(arm, X, samples=self.samples) for arm in arms)
        )

        best_arm_indexes = np.atleast_1d(
            cast(NDArray[np.int_], np.argmax(means, axis=0))
        )

        choice_idx_to_explore = rng.random(size=len(best_arm_indexes)) < self.epsilon

        return [
            arms[cast(int, choice)] if not explore else cast(AT, rng.choice(arms))  # type: ignore
            for explore, choice in zip(choice_idx_to_explore, best_arm_indexes)
        ]


class ThompsonSampling:
    """
    Policy object for Thompson sampling.

    """

    def __repr__(self) -> str:
        return "ThompsonSampling()"

    def __call__(
        self,
        arms: List[Arm[L, T]],
        X: Union[NDArray[np.float_], csc_array],
        rng: np.random.Generator,
    ) -> List[Arm[L, T]]:
        """Choose an arm using Thompson sampling."""

        samples = np.stack(tuple(_draw_one_sample(arm, X) for arm in arms))

        best_arm_indexes = np.atleast_1d(
            cast(NDArray[np.int_], np.argmax(samples, axis=0))
        )

        return [arms[cast(int, choice)] for choice in best_arm_indexes]


class UpperConfidenceBound:
    """
    Policy object for upper confidence bound.

    Parameters
    ----------
    alpha : float, default=0.68
        Confidence level (one-sided)
    samples : int, default=1000
        Number of samples to use for computing the arm upper bounds.

    """

    def __repr__(self) -> str:
        return f"UpperConfidenceBound(alpha={self.alpha}, samples={self.samples})"

    def __init__(self, alpha: float = 0.68, samples: int = 1000):
        self.alpha = alpha
        self.samples = samples

    def __call__(
        self,
        arms: List[Arm[L, T]],
        X: Union[NDArray[np.float_], csc_array],
        rng: np.random.Generator,
    ) -> List[Arm[L, T]]:
        """Choose an arm using upper confidence bound."""

        upper_bounds = np.stack(
            tuple(
                _compute_arm_upper_bound(arm, X, alpha=self.alpha, samples=self.samples)
                for arm in arms
            )
        )

        best_arm_indexes = np.atleast_1d(
            cast(NDArray[np.int_], np.argmax(upper_bounds, axis=0))
        )

        return [arms[cast(int, choice)] for choice in best_arm_indexes]



import os
from enum import Enum
from functools import cached_property
from typing import Any, Tuple, Union, cast

import numpy as np
from attr import dataclass
from numpy.typing import NDArray
from scipy.sparse import csc_array, csc_matrix, csr_matrix, diags, eye, issparse
from scipy.sparse.linalg import splu, spsolve, use_solver
from scipy.stats import Covariance
from scipy.stats._multivariate import _squeeze_output

use_solver(useUmfpack=False)

try:
    from sksparse.cholmod import cholesky as cholmod_cholesky

    use_cholmod = True
except ImportError:
    use_cholmod = False

try:
    import scikits.umfpack as um

    use_umfpack = True
except ImportError:
    use_umfpack = False


if os.environ.get("BB_NO_SUITESPARSE", "0") == "1":
    use_cholmod = False
    use_umfpack = False


class SparseSolver(Enum):
    """Enum for sparse solvers."""

    SUPERLU = 0
    CHOLMOD = 1
    UMFPACK = 2


if use_cholmod:
    solver = SparseSolver.CHOLMOD
elif use_umfpack:
    solver = SparseSolver.UMFPACK
else:
    solver = SparseSolver.SUPERLU


@dataclass
class LUObject:
    L: csr_matrix
    Pr: csc_array


class CovViaSparsePrecision(Covariance):
    """Covariance class for sparse precision matrices.

    Parameters
    ----------
    prec : csc_array
        Sparse precision matrix.
    use_suitesparse : bool, optional
        Whether to use the sksparse.cholmod module for solving linear systems.
        If False, use scipy.sparse.linalg.splu instead. Default is True if
        sksparse.cholmod is installed, False otherwise.

    Raises
    ------
    ValueError
        If prec is not a sparse array.
    ValueError
        If prec is not symmetric.

    Notes
    -----
    This class is used to colorize samples from a standard normal distribution
    with a sparse precision matrix. The precision matrix is assumed to be
    positive-definite, because our Bayesian linear regression models always
    have positive-definite precision matrices (due to the prior).

    A coloring transform from a precision matrix is performed by first finding
    a factor W such that W W^T = P. Then, the colorizing transform is given by
    W^T X = Z, where Z is a standard normal random variable. The resulting
    random variable X has the precision matrix P.

    If using suitesparse, this is done by computing the Cholesky decomposition
    of P, which may or may not be permuted. CHOLMOD's solve_Lt solves W^T X = Z
    to color X with the permuted precision matrix P'. This P' is the precision
    matrix we would have gotten if we'd reordered the features of the training
    data to be in the order given by the permutation. Therefore, we need only
    apply the inverse permutation to X to get samples colored with our actual
    precision matrix P.

    If not using suitesparse, this is done by computing the LU decomposition
    using SuperLU. By telling SuperLU that the matrix is symmetric by setting
    options=dict(SymmetricMode=True), diag_pivot_thresh=0, and
    permc_spec="MMD_AT_PLUS_A", we can prevent partial pivoting and get a
    symmetric factorization L L^T = P', where P' is some permutation of P.
    L^T X = Z is solved to color X with the permuted precision matrix P', same
    as with suitesparse. The inverse permutation is applied to X to get samples
    colored with the original precision matrix P.
    """

    def __init__(self, prec: csc_array, solver=solver):
        if not issparse(prec):
            raise ValueError("prec must be a sparse array")

        self.solver = solver

        self._precision = prec

        if self.solver == SparseSolver.CHOLMOD:
            self._W = cholmod_cholesky(csc_matrix(prec))

        elif self.solver == SparseSolver.UMFPACK:
            umc = um.UmfpackContext()
            # Tells umfpack that we have a symmetric matrix and we only want to pivot on the diagonal
            umc.control[um.UMFPACK_STRATEGY] = um.UMFPACK_STRATEGY_SYMMETRIC  # type: ignore
            umc.control[um.UMFPACK_SYM_PIVOT_TOLERANCE] = 0.0  # type: ignore
            umc.control[um.UMFPACK_SCALE] = um.UMFPACK_SCALE_NONE  # type: ignore

            L, U, P, Q, _, _ = umc.lu(csc_matrix(prec))
            if not (P == Q).all():
                raise ValueError("P != Q. Was a diagonal element of prec zero?")

            self._W = LUObject(
                L=L.dot(diags(np.sqrt(U.diagonal()))),
                Pr=csc_array((np.ones(L.shape[0]), (P, np.arange(L.shape[0])))),
            )

        else:
            # Tells SuperLU that we have a symmetric matrix and we only want to pivot on the diagonal
            splu_ = splu(
                prec,
                diag_pivot_thresh=0,
                permc_spec="MMD_AT_PLUS_A",
                options=dict(SymmetricMode=True),
            )
            if (splu_.perm_r != splu_.perm_c).any():
                raise ValueError("W must be symmetric")

            self._W = LUObject(
                L=splu_.L.dot(diags(np.sqrt(splu_.U.diagonal()))),
                Pr=csc_array(
                    (
                        np.ones(splu_.L.shape[0]),
                        (splu_.perm_r, np.arange(splu_.L.shape[0])),
                    )
                ),
            )

        self._rank = prec.shape[-1]  # must be full rank for cholesky
        self._shape = prec.shape
        self._allow_singular = False

    @property
    def colorize_solve(self) -> Any:
        if self.solver == SparseSolver.CHOLMOD:
            return lambda x: self._W.apply_Pt(  # type: ignore
                self._W.solve_Lt(x, False)  # type: ignore
            )

        elif self.solver == SparseSolver.UMFPACK:
            return lambda x: self._W.Pr @ um.spsolve(self._W.L.T, x)
        return lambda x: self._W.Pr.T @ spsolve(csc_array(self._W.L.T), x)

    @cached_property
    def _covariance(self) -> csc_array:
        return spsolve(self._precision, eye(self._precision.shape[0], format="csc"))

    def _whiten(self, x):
        raise NotImplementedError("Not implemented for sparse matrices")

    def _colorize(self, x: NDArray[np.float_]) -> NDArray[np.float_]:
        samples = self.colorize_solve(x.T).T
        # csc_arrray and csc_matrix have different behavior, so CHOLMOD doesn't
        # auto-squeeze the output. We do it here.
        if samples.shape[0] == 1:
            return samples[0]
        return samples


def multivariate_normal_sample_from_sparse_covariance(
    mean: Union[csc_array, NDArray[np.float_], None],
    cov: Covariance,
    size: int = 1,
    random_state: Union[int, None, np.random.Generator] = None,
) -> NDArray[np.float_]:
    """
    Sample from a multivariate normal distribution with mean mu (default 0)
    and sparse precision matrix Q.

    Parameters
    ----------
    mean : array_like, optional
        Mean of the distribution. Default is 0.
    prec : array_like
        Precision matrix of the distribution. Ideally a csc sparse array.
    size : int or tuple of ints, optional
        Given a shape of, for example, (m,n,k), m*n*k samples are generated,
        and packed in an m-by-n-by-k arrangement. Because each sample is
        N-dimensional, the output shape is (m,n,k,N). If no shape is specified,
        a single (N-D) sample is returned.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random.default_rng`.

    Returns
    -------
    out : ndarray
        The drawn samples, of shape size, if that was provided. If not, the
        shape is (N,).

    """
    # Set the random state
    rng = np.random.default_rng(random_state)

    # Compute size from the shape of Q plus the size parameter
    gen_size = cast(Tuple[int, int], (size,) + (cov.shape[-1],))

    # Sample Z from a standard multivariate normal distribution
    Z = rng.standard_normal(gen_size)

    # Colorize Z
    Y = cast(NDArray[np.float_], cov.colorize(Z))

    # Add the mean vector to each sample if provided
    if mean is not None:
        Y += mean

    return cast(NDArray[np.float_], Y)


def multivariate_t_sample_from_sparse_covariance(
    loc: Union[csc_array, NDArray[np.float_], None],
    shape: Covariance,
    df: float = 1.0,
    size: int = 1,
    random_state: Union[int, None, np.random.Generator] = None,
) -> NDArray[np.float_]:
    """
    Sample from a multivariate t distribution with mean loc, shape matrix
    shape, and degrees of freedom df.

    Parameters
    ----------
    loc : array_like
        Mean of the distribution.
    shape_inv_ : array_like
        Inverse of the shape matrix of the distribution. Ideally a csc sparse array.
    df : int or float, optional
        Degrees of freedom of the distribution. Default is 1.
    size : int or tuple of ints, optional
        Given a shape of, for example, (m,n,k), m*n*k samples are generated,
        and packed in an m-by-n-by-k arrangement. Because each sample is
        N-dimensional, the output shape is (m,n,k,N). If no shape is specified,
        a single (N-D) sample is returned.
    random_state : int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random.default_rng`.

    Returns
    -------
    out : ndarray
        The drawn samples, of shape size, if that was provided. If not, the
        shape is (N,).


    """

    # Set the random state
    rng = np.random.default_rng(random_state)

    x = rng.chisquare(df, size) / df

    z = multivariate_normal_sample_from_sparse_covariance(
        mean=None, cov=shape, size=size, random_state=random_state
    )
    if loc is None:
        loc = np.zeros_like(z)
    samples = loc + z / np.sqrt(x)[..., None]
    samples = _squeeze_output(samples)

    return cast(NDArray[np.float_], samples)



from copy import deepcopy
from dataclasses import Field, dataclass, field
from functools import cached_property, partial
from typing import (
    Any,
    ClassVar,
    Collection,
    Dict,
    Hashable,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from warnings import warn

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.sparse import csc_array, issparse
from sklearn.base import clone
from typing_extensions import Literal, TypeGuard, dataclass_transform, Self

from ._arm import Arm
from ._np_utils import groupby_array
from ._policy_decorators import ArmChoicePolicy
from ._typing import Learner

_B = TypeVar("_B", bound="Bandit")


@dataclass_transform(field_specifiers=(Field[Any], field))
@dataclass
class Bandit:
    """
    Base class for bandits. This class is not meant to be instantiated directly.
    Instead, it should be subclassed and the `learner` and `policy` arguments to
    the `__init_subclass__` method should be set. Optionally, the `delayed_reward`
    argument can be set to `True` to enable delayed rewards.

    Subclass parameters should be passed to `__init_subclass__` during
    subclassing as keyword arguments. These parameters will be added to
    the subclass as class attributes.

    Subclass Parameters
    -------------------

    learner : Learner
        Learner underlying each arm. Must implement `partial_fit` and `sample`.
    policy : Callable[..., Arm]
        Policy for choosing arms. Must take a `Bandit` instance as its first
        argument and return an `Arm` instance.
    delayed_reward : bool, optional
        Whether or not rewards are measured between pulls, by default False.


    Parameters
    ----------
    rng : Union[np.random.Generator, int, None], optional
        Random number generator to use for choosing arms, by default None
    cache : Optional[MutableMapping[Any, str]], optional
        Cache to use for storing arms when `delayed_reward` is set to `True`,
        by default None. If `delayed_reward` is set to `True` and `cache` is
        not set, a `dict` will be used.

    Attributes
    ----------
    learner : Learner
        Learner underlying each arm. Must implement `partial_fit` and `sample`.
    policy : Callable[..., Arm]
        Policy for choosing arms. Must take a `Bandit` instance as its first
        argument and return an `Arm` instance.
    rng: Union[np.random.Generator, None, int]
        Random number generator to use for choosing arms.
    arms: Dict[str, Arm]
        Dictionary of arms.
    last_arm_pulled: Optional[Arm]
        Last arm pulled.
    cache : Optional[MutableMapping[Any, str]]
        Cache to use for storing arms when `delayed_reward` is set to `True`.

    Examples
    --------

    Minimally, a subclass of `Bandit` must pass a `learner` and `policy` to
    `__init_subclass__`. Additionally, all subclasses must define some
    `Arm` instances as class attributes. These will be used to
    initialize the arms of the bandit.

    >>> from bayesianbandits import Arm, GammaRegressor, epsilon_greedy
    >>> clf = GammaRegressor(alpha=1, beta=1)
    >>> policy = epsilon_greedy(0.1)
    >>> def reward_func(x):
    ...     return x
    >>> class MyBandit(Bandit, learner=clf, policy=policy):
    ...     arm1 = Arm("Action 1!", reward_func)
    ...     arm2 = Arm("Action 2!", reward_func)
    >>> bandit = MyBandit(rng=0)

    Once subclassed and instantiated, the `pull` method can be used to pull
    arms. For non-contextual bandits, the `pull` method takes no arguments.
    After pulling an arm, the `update` method can be used to update the
    learner underlying the arm.

    >>> bandit.pull()
    'Action 1!'
    >>> bandit.update(1)

    For delayed reward bandits, the subclass must set the `delayed_reward`
    argument to `__init_subclass__` to `True`.

    >>> class MyDelayedBandit(Bandit, learner=clf, policy=policy, delayed_reward=True):
    ...     arm1 = Arm("Action 1!", reward_func)
    ...     arm2 = Arm("Action 2!", reward_func)
    >>> bandit = MyDelayedBandit(cache={}, rng=0)

    When `delayed_reward` is set to `True`, the `pull` method takes an additional
    argument, `unique_id`, which is used to identify the arm pulled. This is
    used to retrieve the arm from the `cache` when the `update` method is called.

    >>> bandit.pull(unique_id=1)
    'Action 1!'
    >>> bandit.update(1, unique_id=1)

    """

    rng: Union[np.random.Generator, None, int] = field(default=None, repr=False)
    last_arm_pulled: Optional[Arm] = field(default=None, init=False, repr=False)
    cache: Optional[MutableMapping[Any, str]] = field(default=None, repr=False)

    learner: ClassVar[Learner]
    policy: ClassVar[ArmChoicePolicy]

    _delayed_reward: ClassVar[bool] = False

    def __init_subclass__(
        cls,
        /,
        learner: Learner,
        policy: ArmChoicePolicy,
        delayed_reward: bool = False,
        **kwargs: Dict[str, Any],
    ):
        """Initialize a subclass of `Bandit`.

        Parameters
        ----------
        learner : Learner
            Learner to use for each arm.
        policy : Callable[..., Arm]
            Policy to use for choosing arms.
        """
        # Deprecation warning
        warn(
            "The `Bandit` class is deprecated and will be removed in a future "
            "release. Please use `Agent` or `ContextualAgent` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init_subclass__(**kwargs)

        # set learner and policy as class variables
        cls.learner = learner
        cls.policy = staticmethod(policy)  # type: ignore

        # if delayed reward, set cache as an initializable instance variable,
        # otherwise leave it uninitializable
        if delayed_reward is True:
            cls._delayed_reward = True
            setattr(cls, "cache", field(default_factory=dict, repr=False))
        else:
            setattr(cls, "cache", field(default=None, init=False, repr=False))

        # make sure cache is annotated for dataclass magic
        annotations: dict[str, Any] = getattr(cls, "__annotations__", {})
        annotations["cache"] = Optional[MutableMapping[Any, Arm]]

        # collect and annotate all Arm instances
        arm_annotations: Dict[str, Any] = {}

        for name, attr in cls.__dict__.items():
            if isinstance(attr, Arm):
                arm_annotations[name] = Arm
                setattr(
                    cls,
                    name,
                    field(default_factory=partial(deepcopy, attr), init=False),
                )

        # make sure all Arm instances are first in the annotations
        # for dataclass magic - this is unnecessary in Python 3.10
        other_annotations = {
            k: v for k, v in annotations.items() if k not in arm_annotations
        }
        arm_annotations.update(other_annotations)

        setattr(cls, "__annotations__", arm_annotations)

        # modifies cls in-place
        dataclass(cls)

    def arm(self, action_token: Any) -> Arm:
        """Get the arm with the given token.

        Parameters
        ----------
        token : Any
            Token of the arm to retrieve.

        Returns
        -------
        Arm
            Arm with the given token.

        Raises
        ------
        KeyError
            Raised when the token is not in the bandit.
        """
        matching_arms = [
            arm for arm in self.arms.values() if arm.action_token == action_token
        ]
        if len(matching_arms) == 0:
            raise KeyError(f"Arm with action token {action_token} not found.")
        if len(matching_arms) > 1:
            raise KeyError(
                f"Multiple arms with action token {action_token} found. "
                "Please use a unique token."
            )
        return matching_arms[0]

    @overload
    def pull(self, X: Union[ArrayLike, csc_array], /) -> Any:
        ...

    @overload
    def pull(self, X: Union[ArrayLike, csc_array], /, *, unique_id: Any) -> Any:
        ...

    @overload
    def pull(self, /) -> Any:
        ...

    @overload
    def pull(self, /, *, unique_id: Any) -> Any:
        ...

    def pull(
        self,
        X: Union[ArrayLike, csc_array, None] = None,
        /,
        **kwargs: Any,
    ) -> Union[Any, List[Any]]:
        """Choose an arm and pull it. Set `last_arm_pulled` to the name of the
        arm that was pulled.

        Parameters
        ----------
        X : Optional[ArrayLike]
            Context for the bandit. Only provided when the @contextual
            decorator is used.

        Keyword Arguments
        -----------------
        unique_id : Any | Collection[Any]
            Unique identifier for the pull. Required when the `@delayed_reward`
            decorator is used. If a Collection is provided and the bandit is
            contextual, the `X` array must have the same number of rows as the
            length of the Collection. The pull will be performed as a batch and
            the tokens will be returned as a list.

        Returns
        -------
        Any | List[Any]
            Token of the arm pulled. If `unique_id` is a Collection, a list of
            tokens will be returned.

        Raises
        ------
        DelayedRewardException
            Raised when `unique_id` is not provided and `delayed_reward = True`,
            or when any `unique_id` has already been used.
        ValueError
            Raised when `unique_id` is not hashable, or when `X` does not have
            the same number of rows as `unique_id` when `unique_id` is a
            Collection. Also raised when `X` has more than one row when
            `delayed_reward = False`.
        """
        assert isinstance(self.rng, np.random.Generator)  # for the type checker

        X_pull, _ = _validate_arrays(X, None, self._contextual, check_y=False)
        unique_id = None

        if self._delayed_reward is True:
            unique_id: Union[Hashable, Collection[Hashable]] = kwargs.get(
                "unique_id", None
            )

            if unique_id is None:
                raise DelayedRewardException(
                    "The `unique_id` keyword argument is required when the "
                    "`delayed_reward = True`."
                )

            if not (
                isinstance(unique_id, Collection) and not isinstance(unique_id, str)
            ):
                unique_id = [unique_id]

            choices = self._pull_batch_delayed_reward(
                X_pull, cast(Collection[Any], unique_id)
            )

            if len(choices) == 1:
                return choices[0]
            return choices

        return self._pull_batch(X_pull)

    def _pull_batch(self, X_pull: NDArray[np.float_]) -> Any:
        """Makes a single decision and pulls one arm.

        If given, validates that `X_pull` has only one row, as making several
        decisions for multiple context vectors is not possible when
        `delayed_reward = False`.

        Parameters
        ----------
        X_pull : NDArray[np.float_]
            Context vector - must have only one row.

        Returns
        -------
        Any
            Token of the arm pulled.

        Raises
        ------
        ValueError
            Raised when `X_pull` has more than one row.
        """
        assert isinstance(self.rng, np.random.Generator)  # for the type checker

        arms = self.policy(self.arms, X_pull, self.rng)

        if isinstance(arms, Arm):
            arms = [arms]

        ret_vals = [arm.pull() for arm in arms]
        self.last_arm_pulled = arms[-1]

        if len(ret_vals) == 1:
            return ret_vals[0]
        return ret_vals

    def _pull_batch_delayed_reward(
        self, X: NDArray[np.float_], unique_ids: Collection[Any]
    ) -> List[Any]:
        """Makes a batch of decisions and pulls a batch of arms for a delayed
        reward bandit.

        This method validates that `X` has the same number of rows as
        `unique_ids`. It also validates that `unique_ids` is a collection of
        hashable objects and that none of the `unique_ids` have been used
        before. This branch is only taken when `delayed_reward = True` and a
        collection of `unique_ids` is provided.

        Parameters
        ----------
        X : NDArray[np.float_]
            Context array - must have the same number of rows as `unique_ids`.
        unique_ids : Collection[Any]
            Unique identifiers for the pulls.

        Returns
        -------
        List[Any]
            Tokens of the arms pulled, in the same order as `unique_ids`.

        Raises
        ------
        ValueError
            Raised when `X` does not have the same number of rows as
            `unique_ids`.
        DelayedRewardException
            Raised when any of the `unique_ids` have already been used.
        ValueError
            Raised when `unique_ids` is not a collection of hashable objects.
        """
        assert self.cache is not None  # for the type checker
        assert isinstance(self.rng, np.random.Generator)  # for the type checker

        if not _validate_unique_ids(unique_ids):
            raise ValueError(
                "All unique_ids must be hashable. "
                "Please use hashable unique identifiers."
            )

        _already_used_ids = [
            unique_id for unique_id in unique_ids if unique_id in self.cache
        ]
        if len(_already_used_ids) > 0:
            raise DelayedRewardException(
                f"The unique_ids {_already_used_ids} have already been used. "
                "Please use unique identifiers."
            )

        if self._contextual:
            if len(unique_ids) != X.shape[0]:
                raise ValueError(
                    "The number of unique_ids must match the number of rows in `X`."
                )
            arms = self.policy(self.arms, X, self.rng)
        else:
            arms = self.policy(self.arms, X.repeat(len(unique_ids), axis=0), self.rng)

        if isinstance(arms, Arm):
            arms = [arms]

        ret_vals = [arm.pull() for arm in arms]

        for unique_id, arm in zip(unique_ids, arms):
            self.cache[unique_id] = arm.name

        self.last_arm_pulled = arms[-1]

        return ret_vals

    @overload
    def update(self, y: ArrayLike, /) -> None:
        ...

    @overload
    def update(self, y: ArrayLike, /, *, unique_id: Any) -> None:
        ...

    @overload
    def update(self, X: Union[ArrayLike, csc_array], y: ArrayLike, /) -> None:
        ...

    @overload
    def update(
        self, X: Union[ArrayLike, csc_array], y: ArrayLike, /, *, unique_id: Any
    ) -> None:
        ...

    def update(
        self,
        X: Union[ArrayLike, csc_array],
        y: Optional[ArrayLike] = None,
        /,
        **kwargs: Any,
    ) -> None:
        """Update the learner for the last arm pulled.

        Parameters
        ----------
        X: ArrayLike
            Context for the bandit. Only provided when the @contextual
            decorator is used.
        y : ArrayLike
            Outcome for the last arm pulled.

        Raises
        ------
        ValueError
            If no arm has been pulled yet.

        Keyword Arguments
        -----------------
        unique_id : Any | Collection[Any]
            Unique identifier for the pull. Required when the `@delayed_reward`
            decorator is used. If a Collection is provided, all arrays must have
            the same length as the Collection. The update will be performed as a
            batch.
        """
        X_fit, y_fit = _validate_arrays(X, y, self._contextual, check_y=True)

        if self.__class__._delayed_reward is True:
            unique_id: Union[Collection[Any], str, None] = kwargs.get("unique_id", None)
            if unique_id is None:
                raise DelayedRewardException(
                    "The `unique_id` keyword argument is required when the "
                    "`delayed_reward = True`."
                )
            # check if `unique_id` is a non-string iterable
            elif isinstance(unique_id, Collection) and not isinstance(unique_id, str):
                return self._update_batch(
                    X_fit, y_fit, cast(Collection[Any], unique_id)
                )

            try:
                assert self.cache is not None  # this is here for the type checker
                arm = self.cache.pop(unique_id)
            except KeyError:
                raise DelayedRewardException(
                    f"The unique_id {unique_id} is not in the cache. "
                    "Please use a valid unique identifier."
                )
            try:
                arm_to_update = self.arms[arm]
            except KeyError:
                warn(
                    DelayedRewardWarning(
                        f"The arm {arm} is not in the bandit. Skipping."
                    ),
                    stacklevel=2,
                )
                return

        else:
            arm_to_update = cast(Arm, self.last_arm_pulled)

        arm_to_update.update(X_fit, y_fit)

    def _update_batch(
        self, X: NDArray[np.float_], y: NDArray[np.float_], unique_ids: Collection[Any]
    ):
        # fetch the arms names from the cache
        assert self.cache is not None  # for the type checker

        if len(unique_ids) != X.shape[0]:
            raise ValueError(
                "`X`, `y`, and `unique_ids` must have the same number of rows."
            )

        # get indexes of unique_ids that are in the cache
        present_ids = np.array(
            [idx for idx, v in enumerate(unique_ids) if v in self.cache], dtype=int
        )

        # get the arm names from the cache, get None for missing ids
        arm_names = np.array(
            [self.cache.pop(unique_id, None) for unique_id in unique_ids],
            dtype=str,
        )

        # raise if all ids are missing
        if len(present_ids) == 0:
            raise DelayedRewardException(
                f"None of the unique_ids {unique_ids} are in the cache. "
                "Please use valid unique identifiers."
            )

        # warn if some ids are missing
        if (missing_ids := len(arm_names) - len(present_ids)) > 0:
            warn(
                f"{missing_ids} unique_ids not in the cache. Skipping those updates.",
                DelayedRewardWarning,
                stacklevel=2,
            )

        # update the arms, dropping the missing ids
        for X_part, y_part, arms in groupby_array(
            X[present_ids],
            y[present_ids],
            arm_names[present_ids],
            by=arm_names[present_ids],
        ):
            arm_name = arms[0]
            try:
                arm_to_update = self.arms[arm_name]
            except KeyError:
                warn(
                    DelayedRewardWarning(
                        f"The arm {arm_name} is not in the bandit. Skipping."
                    ),
                    stacklevel=2,
                )
                continue

            arm_to_update.update(X_part, y_part)

    @overload
    def sample(self, X: Union[ArrayLike, csc_array], /, *, size: int = 1) -> ArrayLike:
        ...

    @overload
    def sample(self, /, *, size: int = 1) -> ArrayLike:
        ...

    def sample(
        self,
        X: Union[ArrayLike, csc_array, None] = None,
        /,
        *,
        size: int = 1,
        **kwargs: Any,
    ) -> ArrayLike:
        """Sample from the bandit by choosing an arm according to the
        context vector `X`. For each sample, the arm is chosen according
        to the `policy` algorithm and then a sample is drawn from the
        learner.

        Parameters
        ----------
        X : Optional[ArrayLike]
            Context for the bandit. Only provided when the @contextual
            decorator is used.
        size : int, default=1
            Number of samples to draw.

        """
        assert isinstance(self.rng, np.random.Generator)  # for the type checker
        X_sample, _ = _validate_arrays(X, None, self._contextual, check_y=False)
        # choose an arm, draw a sample, and repeat `size` times
        # TODO: this is not the most efficient way to do this
        # but I can't imagine a situation where this would be a bottleneck.
        return np.array(
            [
                cast(Arm, self.policy(self.arms, X_sample, self.rng)).sample(X_sample)
                for _ in range(size)
            ]
        )

    @overload
    def decay(
        self, /, *, decay_rate: Optional[float] = None, decay_last_arm: bool = True
    ) -> None:
        ...

    @overload
    def decay(
        self,
        X: Union[ArrayLike, csc_array],
        /,
        *,
        decay_rate: Optional[float] = None,
        decay_last_arm: bool = True,
    ) -> None:
        ...

    def decay(
        self,
        X: Union[ArrayLike, csc_array, None] = None,
        /,
        *,
        decay_rate: Optional[float] = None,
        decay_last_arm: bool = True,
    ) -> None:
        """Decay the all arms in the bandit.

        Optionally, decay the last arm pulled.

        Parameters
        ----------
        X : Optional[ArrayLike]
            Context for the bandit. Only provided when the @contextual
            decorator is used.
        decay_last_arm : bool, default=True
            Whether to decay the last arm pulled.
        """
        X_decay, _ = _validate_arrays(
            X, None, contextual=self._contextual, check_y=False
        )

        for arm in self.arms.values():
            if decay_last_arm or arm is not self.last_arm_pulled:
                arm.decay(X_decay, decay_rate=decay_rate)

    def __post_init__(self) -> None:
        """Moves all class attributes that are instances of `Arm` to instance
        attributes.

        This ensures that the bandit can be pickled."""

        self.rng = np.random.default_rng(self.rng)
        self._contextual = False

        # initialize the arms with copies of the learner and
        # point the learner rng to the bandit rng
        for arm in self.arms.values():
            self._set_learner(arm)

        if Arm not in self.__annotations__.values():
            raise ValueError(
                "A bandit must have at least one arm. "
                "Add an arm to the class definition."
            )

    def _set_learner(self, arm: Arm) -> None:
        """Set the learner for an arm, if it is not already set."""
        if arm.learner is None:
            arm.learner = clone(self.learner)  # type: ignore
            arm.learner.set_params(random_state=self.rng)

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """Set the state of the bandit.

        This method is called by pickle when loading a bandit from disk.
        This implementation ensures that arms that are no longer defined
        are deleted and arms that are newly defined are initialized, while
        preserving the state of the arms that are still defined.
        """
        self.__dict__.update(state)

        # 0.4.6+ compatibility: add name to arms - this attribute was added in 0.4.6
        # and is not present in older versions, breaking old pickles
        for arm_name, arm in self.arms.items():
            if not hasattr(arm, "name"):
                arm.name = arm_name

        currently_defined_arms = {
            name: attr
            for name, attr in self.__class__.__annotations__.items()
            if attr is Arm
        }

        # delete arms that are no longer defined
        to_del = [
            arm_name for arm_name in self.arms if arm_name not in currently_defined_arms
        ]

        for arm_name in to_del:
            del self.arms[arm_name]
            del self.__dict__[arm_name]

        # initialize arms that are newly defined
        for arm_name, _ in currently_defined_arms.items():
            if arm_name not in self.arms:
                new_arm = self.__class__.__dataclass_fields__[
                    arm_name
                ].default_factory()
                self._set_learner(new_arm)
                setattr(self, arm_name, new_arm)  # type: ignore
                self.arms[arm_name] = self.__dict__[arm_name]

    @cached_property
    def arms(self: Self) -> Dict[str, Arm]:
        return {
            name: attr for name, attr in self.__dict__.items() if isinstance(attr, Arm)
        }


@overload
def _validate_arrays(
    X: Union[ArrayLike, csc_array],
    y: Optional[ArrayLike],
    /,
    contextual: bool,
    check_y: Literal[True] = True,
) -> Tuple[NDArray[np.float_], NDArray[np.float_]]:
    ...


@overload
def _validate_arrays(
    X: Union[ArrayLike, csc_array, None],
    y: Literal[None],
    /,
    contextual: bool,
    check_y: Literal[False] = False,
) -> Tuple[NDArray[np.float_], None]:
    ...


def _validate_unique_id(
    unique_id: Any,
) -> TypeGuard[Hashable]:
    return isinstance(unique_id, Hashable)


def _validate_unique_ids(
    unique_ids: Collection[Any],
) -> TypeGuard[Collection[Hashable]]:
    for unique_id in unique_ids:
        if not _validate_unique_id(unique_id):
            return False
    return True


def _validate_arrays(
    X: Union[ArrayLike, csc_array, None],
    y: Optional[ArrayLike],
    /,
    contextual: bool,
    check_y: bool = True,
) -> Tuple[Union[NDArray[np.float_], csc_array], Optional[NDArray[np.float_]]]:
    """Validate the `X` and `y` arrays.

    Parameters
    ----------
    X : ArrayLike
        Context for the bandit. Only provided when the @contextual
        decorator is used. Otherwise, this position is used for `y`.
    y : Optional[ArrayLike]
        Reward for the bandit. Only provided when the @contextual
        decorator is used. Otherwise, this position should be None.
    contextual : bool
        Whether the bandit is contextual.
    check_y : bool, default=True
        Whether to check the `y` array.

    Returns
    -------
    Tuple[NDArray, NDArray]
        Validated `X` and `y` arrays.
    """
    if check_y:
        should_not_be_none_if_contextual = y
    else:
        should_not_be_none_if_contextual = X

    if contextual and should_not_be_none_if_contextual is None:
        raise ValueError("Context must be provided for a contextual bandit.")
    if not contextual and should_not_be_none_if_contextual is not None:
        raise ValueError("Context must be None for a non-contextual bandit.")

    if contextual:
        X = np.atleast_2d(cast(ArrayLike, X)) if not issparse(X) else X
        y = np.atleast_1d(cast(ArrayLike, y)) if check_y else None
    else:
        y = np.atleast_1d(cast(ArrayLike, X)) if check_y else None
        X = (  # type: ignore
            np.ones_like(y, dtype=float)[:, np.newaxis]
            if check_y
            else np.array([[1]], dtype=float)
        )

    if check_y:
        if X.shape[0] != y.shape[0]:  # type: ignore
            raise ValueError(
                "The number of rows in `X` must match the number of rows in `y`."
            )
    assert isinstance(X, (np.ndarray, csc_array))  # for the type checker
    return X, y


def contextual(
    cls: Type[_B],
) -> Type[_B]:
    """Decorator for making a bandit contextual.

    This decorator adds methods to a Bandit subclass that allow it to be used
    in a contextual setting. The `pull` and `sample` methods will take `X`
    arguments, and the `update` method will take `X` and `y`.

    Parameters
    ----------
    cls : Bandit
        Bandit class to make contextual.

    Returns
    -------
    Bandit
        Contextual bandit class.

    Examples
    --------

    A contextual bandit can be created by decorating a Bandit subclass with
    the `contextual` decorator.

    >>> from bayesianbandits import Arm, GammaRegressor, epsilon_greedy
    >>> clf = GammaRegressor(alpha=1, beta=1)
    >>> policy = epsilon_greedy(0.1)
    >>> def reward_func(x):
    ...     return x
    >>> @contextual
    ... class MyBandit(Bandit, learner=clf, policy=policy):
    ...     arm1 = Arm("Action 1!", reward_func)
    ...     arm2 = Arm("Action 2!", reward_func)
    >>> bandit = MyBandit(rng=0)

    The `pull`, `sample`, and `update` methods now take an `X` argument.

    >>> bandit.pull(1)
    'Action 1!'
    >>> bandit.update(1, 1)

    """

    check_is_bandit(cls)

    orig_post_init = cls.__post_init__

    def _contextual_post_init(self: _B) -> None:
        orig_post_init(self)
        self._contextual = True  # type: ignore

    setattr(cls, "__post_init__", _contextual_post_init)

    return cls


def restless(
    cls: Type[_B],
) -> Type[_B]:
    """Decorator for restless bandits.

    This decorator ensures that the `decay` method of each unselected
    arm is called at each update. The specific behavior of the `decay`
    method is left to the learner.

    """

    check_is_bandit(cls)

    orig_update = cls.update  # type: ignore

    def _restless_update(
        self: Bandit,
        X: ArrayLike,
        y: Optional[ArrayLike] = None,
        **kwargs: ArrayLike,
    ) -> None:
        orig_update(self, X, y, **kwargs)  # type: ignore

        if self._contextual:  # type: ignore
            self.decay(X, decay_last_arm=False)
        else:
            self.decay(decay_last_arm=False)

    setattr(cls, "update", _restless_update)

    return cls


def check_is_bandit(cls: type):
    if not issubclass(cls, Bandit):
        raise ValueError("This decorator can only be used on a Bandit subclass.")


class DelayedRewardException(Exception):
    """Exception raised when the user does not handle delayed reward bandits
    correctly.

    For example, if the user tries to reuse a `unique_id`."""

    pass


class DelayedRewardWarning(UserWarning):
    """Warning raised when the user does not handle delayed reward bandits
    correctly.

    For example, if the user tries to reuse a `unique_id`."""

    pass



from __future__ import annotations

from typing import Any, Generator, Tuple, TypeVar

import numpy as np
from numpy.typing import NDArray

_T = TypeVar("_T", bound=Any)


def groupby_array(
    *arrays: NDArray[_T], by: NDArray[Any]
) -> Generator[Tuple[NDArray[_T], ...], None, None]:
    """Group arrays by a given array.

    Parameters
    ----------
    *arrays : array-like
        Arrays to be grouped.
    by : array-like
        Array to group by.

    Yields
    ------
    array-like
        Grouped arrays.

    Examples
    --------
    >>> import numpy as np
    >>> from bayesianbandits._np_utils import groupby_array
    >>> X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> y = np.array([1, 2, 3])
    >>> for group in groupby_array(X, y, by=y):
    ...     print(group)
    (array([[1, 2, 3]]), array([1]))
    (array([[4, 5, 6]]), array([2]))
    (array([[7, 8, 9]]), array([3]))

    """
    sort_keys = np.argsort(by, kind="stable")
    sorted_by = by[sort_keys]
    sorted_arrays = [array[sort_keys] for array in arrays]

    group_indexes = np.unique(sorted_by, return_index=True)[1][1:]
    split_indexes = np.split(np.arange(len(sorted_by)), group_indexes)

    for split in split_indexes:
        yield tuple(array[split] for array in sorted_arrays)



from __future__ import annotations

from typing import Callable, Dict, List, Union, cast

import numpy as np
from numpy.typing import NDArray
from scipy.sparse import csc_array

from ._arm import Arm

ArmChoicePolicy = Callable[
    [Dict[str, Arm], NDArray[np.float_], np.random.Generator],
    Union[Arm, List[Arm]],
]


def epsilon_greedy(
    epsilon: float = 0.1,
    *,
    samples: int = 1000,
) -> ArmChoicePolicy:
    """Creates an epsilon-greedy choice algorithm. To be used with the
    `Bandit` class.

    Parameters
    ----------
    epsilon : float, default=0.1
        Probability of choosing a random arm.
    samples : int, default=1000
        Number of samples to use to compute the mean of the posterior.

    Returns
    -------
    Callable[[BanditProtocol, NDArray[np.float_]], Arm]
        Closure that chooses an arm using epsilon-greedy.
    """

    def _choose_arm(
        arms: Dict[str, Arm],
        X: NDArray[np.float_],
        rng: np.random.Generator,
    ) -> Union[Arm, List[Arm]]:
        """Choose an arm using epsilon-greedy."""

        arm_list = list(arms.values())

        means = np.stack(
            tuple(_compute_arm_mean(arm, X, samples=samples) for arm in arm_list)
        )

        choices = _return_based_on_size(arm_list, means)

        if not isinstance(choices, list):
            if rng.random() < epsilon:
                return rng.choice(arm_list)  # type: ignore
            else:
                return choices

        choice_idx_to_explore = rng.random(size=len(choices)) < epsilon

        final_choices: List[Arm] = []
        for explore, choice in zip(choice_idx_to_explore, choices):
            if explore:
                final_choices.append(rng.choice(arm_list))  # type: ignore
            else:
                final_choices.append(choice)

        return final_choices

    return _choose_arm


def upper_confidence_bound(
    alpha: float = 0.68,
    *,
    samples: int = 1000,
) -> ArmChoicePolicy:
    """Creates a UCB choice algorithm. To be used with the
    `Bandit` class.

    Actually uses the upper bound of the one-sided credible interval,
    which will deviate from the upper bound of the one-sided confidence
    interval depending on the strength of the prior.

    Parameters
    ----------
    alpha : float, default=0.68
        Upper bound of the one-sided prediction interval.
    samples : int, default=1000
        Number of samples to use to compute the upper bound of the
        credible interval. The larger `alpha` is, the larger `samples`
        should be.

    Returns
    -------
    Callable[[BanditProtocol, NDArray[np.float_]], Arm]
        Closure that chooses an arm using UCB.

    Notes
    -----
    A deeper analysis of this Bayesian UCB algorithm can be found in
    [1].

    References
    ----------
    [1] E. Kaufmann, O. Cappé, and A. Garivier, “On Bayesian upper confidence
        bounds for bandit problems,” In Proceedings of the 15th International
        Conference on Artificial Intelligence and Statistics, 2012.

    """

    if not 0 < alpha < 1:
        raise ValueError("alpha must be in (0, 1).")

    def _choose_arm(
        arms: Dict[str, Arm],
        X: NDArray[np.float_],
        rng: np.random.Generator,
    ) -> Union[Arm, List[Arm]]:
        """Choose an arm using UCB1."""

        arm_list = list(arms.values())

        # Compute the upper bound of the one-sided credible interval for
        # each arm.
        upper_bounds = np.stack(
            tuple(
                _compute_arm_upper_bound(arm, X, alpha=alpha, samples=samples)
                for arm in arm_list
            )
        )

        # Return the arm(s) with the largest upper bound.
        return _return_based_on_size(arm_list, upper_bounds)

    return _choose_arm


def thompson_sampling() -> ArmChoicePolicy:
    """Creates a Thompson sampling choice algorithm. To be used with the
    `Bandit` class.

    Returns
    -------
    Callable[[BanditProtocol, NDArray[np.float_]], Arm]
        Closure that chooses an arm using Thompson sampling.

    Notes
    -----
    This is a very simple implementation of Thompson sampling, identical
    to the one used in [1]. This generally works well in practice and
    has no hyperparameters to tune.

    References
    ----------
    [1] D. Russo, B. Van Row, A. Kazerouni, I. Osband, and Z. Wen, “A
        Tutorial on Thompson Sampling,” Foundations and Trends® in Machine
        Learning, vol. 11, no. 1, pp. 1-96, 2018.
    """

    def _choose_arm(
        arms: Dict[str, Arm],
        X: NDArray[np.float_],
        rng: np.random.Generator,
    ) -> Union[Arm, List[Arm]]:
        """Choose an arm using Thompson sampling."""

        arm_list = list(arms.values())

        # Sample from the posterior distribution for each arm.
        posterior_summaries = np.stack(
            tuple(_draw_one_sample(arm, X) for arm in arm_list)
        )

        return _return_based_on_size(arm_list, posterior_summaries)

    return _choose_arm


def _return_based_on_size(
    arm_list: List[Arm],
    posterior_summaries: NDArray[np.float_],
):
    best_arm_indexes = cast(
        NDArray[np.int_], np.atleast_1d(np.argmax(posterior_summaries, axis=0))
    )

    if len(best_arm_indexes) == 1:
        return arm_list[best_arm_indexes.item()]

    else:
        return [arm_list[cast(int, i)] for i in best_arm_indexes]


def _draw_one_sample(
    arm: Arm,
    X: Union[NDArray[np.float_], csc_array],
) -> NDArray[np.float_]:
    """Draw one sample from the posterior distribution for the arm."""
    return arm.sample(X, size=1).squeeze(axis=0)


def _compute_arm_upper_bound(
    arm: Arm,
    X: Union[NDArray[np.float_], csc_array],
    *,
    alpha: float = 0.68,
    samples: int = 1000,
) -> np.float_:
    """Compute the upper bound of a one-sided credible interval with size
    `alpha` from the posterior distribution for the arm."""
    posterior_samples = arm.sample(X, size=samples)

    return np.quantile(posterior_samples, q=alpha, axis=0)  # type: ignore


def _compute_arm_mean(
    arm: Arm,
    X: Union[NDArray[np.float_], csc_array],
    *,
    samples: int = 1000,
) -> np.float_:
    """Compute the mean of the posterior distribution for the arm."""
    posterior_samples = arm.sample(X, size=samples)
    return np.mean(posterior_samples, axis=0)
